/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.oozie.action.hadoop;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.hive.cli.CliDriver;

public class HiveMain extends LauncherMain {

    private static final String HIVE_L4J_PROPS = "hive-log4j.properties";
    private static final String HIVE_EXEC_L4J_PROPS = "hive-exec-log4j.properties";
    private static final String HIVE_SITE_CONF = "hive-site.xml";
    private static final String HIVE_SCRIPT = "oozie.hive.script";
    private static final String HIVE_PARAMS = "oozie.hive.params";

    public static void main(String[] args) throws Exception {
        run(HiveMain.class, args);
    }

    private Configuration initActionConf() {
        // Loading action conf prepared by Oozie
        Configuration hiveConf = new Configuration(false);

        String actionXml = System.getProperty("oozie.action.conf.xml");

        if (actionXml == null) {
            throw new RuntimeException("Missing Java System Property [oozie.action.conf.xml]");
        }
        if (!new File(actionXml).exists()) {
            throw new RuntimeException("Action Configuration XML file [" + actionXml + "] does not exist");
        } else {
            System.out.println("Using action configuration file " + actionXml);
        }

        hiveConf.addResource(new Path("file:///", actionXml));

        // Propagate delegation related props from launcher job to Hive job
        if (System.getenv("HADOOP_TOKEN_FILE_LOCATION") != null) {
            hiveConf.set("mapreduce.job.credentials.binary",
                    System.getenv("HADOOP_TOKEN_FILE_LOCATION"));
            System.out.println("------------------------");
            System.out.println("Setting env property for mapreduce.job.credentials.binary to:"
                    + System.getenv("HADOOP_TOKEN_FILE_LOCATION"));
            System.out.println("------------------------");
            System.setProperty("mapreduce.job.credentials.binary",
                    System.getenv("HADOOP_TOKEN_FILE_LOCATION"));
        } else {
            System.out.println("Non-Kerberos execution");
        }

        // Have to explicitly unset this property or Hive will not set it.
        hiveConf.set("mapred.job.name", "");

        // See https://issues.apache.org/jira/browse/HIVE-1411
        hiveConf.set("datanucleus.plugin.pluginRegistryBundleCheck", "LOG");

        return hiveConf;
    }

    protected void run(String[] args) throws Exception {
        System.out.println();
        System.out.println("Oozie Hive action configuration");
        System.out.println("=================================================================");

        Configuration hiveConf = initActionConf();

        // Write the action configuration out to hive-site.xml
        OutputStream os = new FileOutputStream(HIVE_SITE_CONF);
        hiveConf.writeXml(os);
        os.close();

        System.out.println();
        System.out.println("Hive Configuration Properties:");
        System.out.println("------------------------");
        for (Entry<String, String> entry : hiveConf) {
            System.out.println(entry.getKey() + "=" + entry.getValue());
        }
        System.out.flush();
        System.out.println("------------------------");
        System.out.println();

        List<String> arguments = new ArrayList<String>();
        String scriptPath = hiveConf.get(HIVE_SCRIPT);

        if (scriptPath == null) {
            throw new RuntimeException("Action Configuration does not have [oozie.hive.script] property");
        }

        if (!new File(scriptPath).exists()) {
            throw new RuntimeException("Hive script file [" + scriptPath + "] does not exist");
        }

        //Logfile to capture job IDs
        String hadoopJobId = System.getProperty("oozie.launcher.job.id");
        if (hadoopJobId == null) {
            throw new RuntimeException("Launcher Hadoop Job ID system property not set");
        }

        String logFile = new File("hive-oozie-" + hadoopJobId + ".log").getAbsolutePath();

        // Preparing log4j configuration
        URL log4jFile = Thread.currentThread().getContextClassLoader().getResource("log4j.properties");
        if (log4jFile != null) {

            String hiveLogLevel = hiveConf.get("oozie.hive.log.level", "INFO");

            // append required Hive properties to the default hadoop log4j file
            Properties hadoopProps = new Properties();
            hadoopProps.load(log4jFile.openStream());

            hadoopProps.setProperty("log4j.root.logger", hiveLogLevel + ", DRFA, EventCounter");

            hadoopProps.setProperty("log4j.logger.org.apache.hadoop.hive", "INFO, jobid");
            
            hadoopProps.setProperty("log4j.appender.jobid", "org.apache.log4j.FileAppender");
            hadoopProps.setProperty("log4j.appender.jobid.file", logFile);
            hadoopProps.setProperty("log4j.appender.jobid.layout", "org.apache.log4j.PatternLayout");
            hadoopProps.setProperty("log4j.appender.jobid.layout.ConversionPattern", "%-4r [%t] %-5p %c %x - %m%n");

            hadoopProps.setProperty("log4j.appender.DRFA", "org.apache.log4j.ConsoleAppender");
            hadoopProps.setProperty("log4j.appender.DRFA.layout", "org.apache.log4j.PatternLayout");
            hadoopProps.setProperty("log4j.appender.DRFA.layout.ConversionPattern", "%d{ISO8601} %p %c: %m%n");

            hadoopProps.setProperty("log4j.appender.EventCounter", "org.apache.hadoop.metrics.jvm.EventCounter");

            hadoopProps.setProperty("log4j.category.DataNucleus", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.Datastore", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.Datastore.Schema", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.JPOX.Datastore", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.JPOX.Plugin", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.JPOX.MetaData", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.JPOX.Query", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.JPOX.General", "ERROR,DRFA");
            hadoopProps.setProperty("log4j.category.JPOX.Enhancer", "ERROR,DRFA");

            String localProps = new File(HIVE_L4J_PROPS).getAbsolutePath();
            OutputStream os1 = new FileOutputStream(localProps);
            hadoopProps.store(os1, "");
            os1.close();

            localProps = new File(HIVE_EXEC_L4J_PROPS).getAbsolutePath();
            os1 = new FileOutputStream(localProps);
            hadoopProps.store(os1, "");
            os1.close();

            // print out current directory
            File localDir = new File(localProps).getParentFile();
            System.out.println("Current (local) dir = " + localDir.getAbsolutePath());
        }
        
        // Prepare the Hive Script
        String script = readStringFromFile(scriptPath);
        System.out.println("Original script [" + scriptPath + "] content: ");
        System.out.println("------------------------");
        System.out.println(script);
        System.out.println("------------------------");
        System.out.println();

        Map<String, String> varMap = new HashMap<String, String>();

        for (String param : MapReduceMain.getStrings(hiveConf, HIVE_PARAMS)) {
            System.out.println("Param Expression: '" + param + "'");

            int idx = param.indexOf('=');
            if (idx == -1) {
                throw new RuntimeException("Parameter expression must contain an assignment: " + param);
            } else if (idx == 0) {
                throw new RuntimeException("Parameter value not specified: " + param);
            }
            String var = param.substring(0, idx);
            String val = param.substring(idx + 1, param.length());
            System.out.println("var = " + var + ", val = " + val);
            varMap.put(var, val);
        }

        if (varMap.size() > 0) {
            System.out.println("Substituting parameter values...");
            String resolvedScript = substitute(varMap, script);
            scriptPath = scriptPath + ".sub";
            writeStringToFile(scriptPath, resolvedScript);

            System.out.println("Substituted script [" + scriptPath + "] content: ");
            System.out.println("------------------------");
            System.out.println(resolvedScript);
            System.out.println("------------------------");
            System.out.println();
        }

        arguments.add("-f");
        arguments.add(scriptPath);


        System.out.println("Hive command arguments :");
        for (String arg : arguments) {
            System.out.println("             " + arg);
        }

        System.out.println("=================================================================");
        System.out.println();
        System.out.println(">>> Invoking Hive command line now >>>");
        System.out.println();
        System.out.flush();

        try {
            runHive(arguments.toArray(new String[arguments.size()]));
        }
        catch (SecurityException ex) {
            if (LauncherSecurityManager.getExitInvoked()) {
                if (LauncherSecurityManager.getExitCode() != 0) {
                    throw ex;
                }
            }
        }

        System.out.println("\n<<< Invocation of Hive command completed <<<\n");

        // harvesting and recording Hadoop Job IDs
        Properties jobIds = getHadoopJobIds(logFile);
        File file = new File(System.getProperty("oozie.action.output.properties"));
        os = new FileOutputStream(file);
        jobIds.store(os, "");
        os.close();
        System.out.println(" Hadoop Job IDs executed by Hive: " + jobIds.getProperty("hadoopJobs"));
        System.out.println();
    }

    private void runHive(String[] args) throws Exception {
        CliDriver.main(args);
    }

    public static void setHiveScript(Configuration conf, String script, String[] params) {
        conf.set(HIVE_SCRIPT, script);
        MapReduceMain.setStrings(conf, HIVE_PARAMS, params);
    }

    private static String readStringFromFile(String filePath) throws IOException {
        String line;
        BufferedReader br = new BufferedReader(new FileReader(filePath));
        StringBuilder sb = new StringBuilder();
        String sep = System.getProperty("line.separator");
        while ((line = br.readLine()) != null) {
            sb.append(line).append(sep);
        }
        return sb.toString();
     }

    private static void writeStringToFile(String filePath, String str) throws IOException {
        BufferedWriter out = new BufferedWriter(new FileWriter(filePath));
        out.write(str);
        out.close();
    }

    static String substitute(Map<String, String> vars, String expr) {
        for (Map.Entry<String, String> entry : vars.entrySet()) {
            String var = "${" + entry.getKey() + "}";
            String value = entry.getValue();
            expr = expr.replace(var, value);
        }
        return expr;
    }

    private static final String JOB_ID_LOG_PREFIX = "Ended Job = ";

    public static Properties getHadoopJobIds(String logFile) throws IOException {
        Properties props = new Properties();
        StringBuffer sb = new StringBuffer(100);
        if (new File(logFile).exists()) {
            System.err.println("hive log file: " + logFile + "  not present. Therefore no Hadoop jobids found");
            props.setProperty("hadoopJobs", "");
        }
        else {
            BufferedReader br = new BufferedReader(new FileReader(logFile));
            String line = br.readLine();
            String separator = "";
            while (line != null) {
                if (line.contains(JOB_ID_LOG_PREFIX)) {
                    int jobIdStarts = line.indexOf(JOB_ID_LOG_PREFIX) + JOB_ID_LOG_PREFIX.length();
                    String jobId = line.substring(jobIdStarts).trim();
                    sb.append(separator).append(jobId);
                    separator = ",";
                }
                line = br.readLine();
            }
            br.close();
            props.setProperty("hadoopJobs", sb.toString());
        }
        return props;
    }

}
