/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.oozie.action.hadoop;

import java.io.IOException;
import java.io.StringReader;
import java.util.List;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileStatus;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.oozie.action.ActionExecutorException;
import org.apache.oozie.client.WorkflowAction;
import org.apache.oozie.util.XConfiguration;
import org.apache.oozie.util.XmlUtils;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;

import static org.apache.oozie.action.hadoop.SqoopMain.SQOOP_DEFAULT_CONF;
import static org.apache.oozie.action.hadoop.SqoopMain.SQOOP_SITE_CONF;
import static org.apache.oozie.action.hadoop.LauncherMapper.CONF_OOZIE_ACTION_MAIN_CLASS;

public class SqoopActionExecutor extends JavaActionExecutor {


    public SqoopActionExecutor() {
        super("sqoop");
    }

    @Override
    protected List<Class> getLauncherClasses() {
        List<Class> classes = super.getLauncherClasses();
        classes.add(LauncherMain.class);
        classes.add(MapReduceMain.class);
        classes.add(SqoopMain.class);
        return classes;
    }

    @Override
    protected String getLauncherMain(Configuration launcherConf, Element actionXml) {
        return launcherConf.get(CONF_OOZIE_ACTION_MAIN_CLASS, SqoopMain.class.getName());
    }

    @Override
    void injectActionCallback(Context context, Configuration launcherConf) {
    }

    protected void addResourceToCache(Configuration conf, FileSystem fs, Path resourcePath)
    throws ActionExecutorException {
        try {
            if (!fs.exists(resourcePath) || !fs.isFile(resourcePath)) {
                return;
            }
            Path parentPath = resourcePath.getParent();
            String resourceName = resourcePath.getName();
            addToCache(conf, parentPath, resourceName + "#" + resourceName, false);
        } catch (Exception ex) {
            throw convertException(ex);
        }
    }

    @Override
    Configuration setupLauncherConf(Configuration conf, Element actionXml, Path appPath, Context context)
            throws ActionExecutorException {
        try {
            super.setupLauncherConf(conf, actionXml, appPath, context);
            Namespace ns = actionXml.getNamespace();

            // Look for hive-log4j.properties and hive-exec-log4j.properties
            // in the configuration directory and add them to the distributed
            // cache.
            Element e = actionXml.getChild("conf-dir", ns);
            if (e != null) {
                Path hiveConfPath = new Path(appPath, e.getTextTrim());
                FileSystem fs = getActionFileSystem(context, actionXml);
                FileStatus status = fs.getFileStatus(hiveConfPath);
                if (status.isDir()) {
//                    addResourceToCache(conf, fs, new Path(hiveConfPath,HIVE_L4J_PROPS));
//                    addResourceToCache(conf, fs, new Path(hiveConfPath, HIVE_EXEC_L4J_PROPS));
                }
            }
            return conf;
        } catch (IOException ex) {
            throw convertException(ex);
        }
    }

    @Override
    @SuppressWarnings("unchecked")
    Configuration setupActionConf(Configuration actionConf, Context context, Element actionXml, Path appPath)
            throws ActionExecutorException {
        super.setupActionConf(actionConf, context, actionXml, appPath);
        Namespace ns = actionXml.getNamespace();

        // Read sqoop-default.xml, sqoop-site.xml, hive-default.xml, hive-site.xml,
        // and any configuration properties defined in the workflow and overlay
        // these properties onto the action configuration in this order.
        Element e = actionXml.getChild("conf-dir", ns);
        try {
            if (e != null) {
                Path sqoopConfPath = new Path(appPath, e.getTextTrim());
                FileSystem fs = getActionFileSystem(context, actionXml);
                FileStatus status = fs.getFileStatus(sqoopConfPath);
                if (status.isDir()) {
                    loadConfiguration(actionConf, fs, new Path(sqoopConfPath, SQOOP_DEFAULT_CONF));
                    loadConfiguration(actionConf, fs, new Path(sqoopConfPath, SQOOP_SITE_CONF));
//                    loadConfiguration(actionConf, fs, new Path(sqoopConfPath, HIVE_DEFAULT_CONF));
//                    loadConfiguration(actionConf, fs, new Path(sqoopConfPath, HIVE_SITE_CONF));
                }
            }

            e = actionXml.getChild("configuration", ns);
            if (e != null) {
                String strConf = XmlUtils.prettyPrint(e).toString();
                XConfiguration inlineConf = new XConfiguration(new StringReader(strConf));
                checkForDisallowedProps(inlineConf, "inline configuration");
                XConfiguration.copy(inlineConf, actionConf);
            }
        } catch (IOException ex) {
            throw convertException(ex);
        }

        String command = actionXml.getChild("command", ns).getTextTrim();

        SqoopMain.setSqoopCommand(actionConf, command);
        return actionConf;
    }

    private void loadConfiguration(Configuration baseConf, FileSystem fs, Path confPath)
    throws ActionExecutorException {
        try {
            if (!fs.exists(confPath) || !fs.isFile(confPath)) {
                return;
            }
            Configuration newConf = new XConfiguration(fs.open(confPath));
            checkForDisallowedProps(newConf, confPath.getName());
            XConfiguration.copy(newConf, baseConf);
        } catch (IOException ex) {
            throw convertException(ex);
        }
    }

    @Override
    protected boolean getCaptureOutput(WorkflowAction action) throws JDOMException {
        return true;
    }
}
