/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.oozie.action.hadoop;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import com.cloudera.sqoop.Sqoop;

public class SqoopMain extends LauncherMain {

    public static final String SQOOP_DEFAULT_CONF = "sqoop-default.xml";
    public static final String SQOOP_SITE_CONF = "sqoop-site.xml";

    public static void main(String[] args) throws Exception {
        run(SqoopMain.class, args);
    }

    protected void run(String[] args) throws Exception {
        System.out.println();
        System.out.println("Oozie Sqoop action configuration");
        System.out.println("=================================================================");

        // loading action conf prepared by Oozie
        Configuration actionConf = new Configuration(false);

        String actionXml = System.getProperty("oozie.action.conf.xml");

        if (actionXml == null) {
            throw new RuntimeException("Missing Java System Property [oozie.action.conf.xml]");
        }
        if (!new File(actionXml).exists()) {
            throw new RuntimeException("Action Configuration XML file [" + actionXml + "] does not exist");
        }

        actionConf.addResource(new Path("file:///", actionXml));

        // See http://issues.apache.org/jira/browse/HIVE-1411
        actionConf.set("datanucleus.plugin.pluginRegistryBundleCheck", "LOG");

        // Propagate delegation related props from launcher job to Hive job
        if (System.getenv("HADOOP_TOKEN_FILE_LOCATION") != null) {
            actionConf.set("mapreduce.job.credentials.binary",
                    System.getenv("HADOOP_TOKEN_FILE_LOCATION"));
            System.out.println("------------------------");
            System.out.println("Setting env property for mapreduce.job.credentials.binary to:"
                    + System.getenv("HADOOP_TOKEN_FILE_LOCATION"));
            System.out.println("------------------------");
            System.setProperty("mapreduce.job.credentials.binary",
                    System.getenv("HADOOP_TOKEN_FILE_LOCATION"));
        } else {
            System.out.println("Non-Kerberos execution");
        }

        // Write the action configuration out to sqoop-site.xml
        // AND hive-site.xml as properties contained in the configuration
        // can target either Sqoop proper or Hive (in the event
        // of a Sqoop hive-import operation.
        OutputStream os = new FileOutputStream(SQOOP_SITE_CONF);
        actionConf.writeXml(os);
        os.close();

//        os = new FileOutputStream(HIVE_SITE_CONF);
//        actionConf.writeXml(os);
//        os.close();
//

        String command = actionConf.get("oozie.sqoop.command");
        if (command == null) {
            throw new RuntimeException("Action Configuration does not have [oozie.sqoop.command] property");
        }

        String[] sqoopArgs = command.split(" ");

        System.out.println("Sqoop command arguments :");
        for (String arg : sqoopArgs) {
            System.out.println("             " + arg);
        }

        System.out.println("=================================================================");
        System.out.println();
        System.out.println(">>> Invoking Sqoop command line now >>>");
        System.out.println();
        System.out.flush();

        runSqoopJob(sqoopArgs);

        System.out.println();
        System.out.println("<<< Invocation of Sqoop command completed <<<");
        System.out.println();


    }

    protected void runSqoopJob(String[] args) throws Exception {
        // running as from the command line
        Sqoop.main(args);
    }

    public static void setSqoopCommand(Configuration conf, String command) {
        conf.set("oozie.sqoop.command", command);
    }
}
